<?php
namespace Admin\Controller;

use Common\Controller\AdminController;
use Niaoyun\Upgrade\File\FileHelper;
use EB\Api\EpowerFile;
use EB\Tar\TarReader;
class UpgradeController extends AdminController
{
	/** @var  FileHelper */
	private $fileHelper;

	private $dataDir;


	public function _initialize()
	{
		parent::_initialize();


		$this->dataDir = REAL_PATH . "/datas/upgrade";
		$this->fileHelper = new FileHelper();
	}

	private function fail($text)
	{
		$data = array('result' => false, 'text' => $text);

		return $data;
	}

	private function actionVerify($action)
	{
		$actions = array(
			'info',
			'check',
			'connect',
			'download',
			'showfile',
		);

		if (!in_array($action, $actions, true)) {
			$this->ajaxReturn($this->fail('参数异常'));
		}
	}

	public function index()
	{

		$action = I('get.action');
		$version = I('get.version');
		if (IS_GET && empty($action)) {
			if (!$this->fileHelper->file_exists($this->dataDir)) {
				
				try {

					$this->fileHelper->mkdirIfNotExists($this->dataDir);
				} catch (\Exception $e) {
					$this->error('目录['.htmlentities($this->dataDir, ENT_QUOTES).']不存在,请创建并确保程序有写入权限!!!' , '', 30);
				}
			}
			//检查写入权限
			if (!$this->fileHelper->writeable($this->dataDir)) {
				$this->error('目录['.htmlentities($this->dataDir, ENT_QUOTES).']不可写入,请修改目录权限!!!' , '', 30);
			}
			$version_file = $this->dataDir . "/data/version.json";
			if(!is_file($version_file)){
				$this->error('版本文件不完整！' , '', 30);
			}
			$json = file_get_contents($version_file);
			$info=json_decode($json,true);
			if(!$info){
				$this->error('版本文件不完整！' , '', 30);
			}
			
			$this->assign('versionInfo', $info);
			$this->display();
		} else {
			$this->actionVerify($action);
			call_user_func('self::' . $action, $version);
		}
	}

	private function check($version) {
		
		$EpowerFile=new EpowerFile();
		$result=$EpowerFile->check($version);
		$this->ajaxReturn($result);
	}


	private function info($version)
	{
		$EpowerFile=new EpowerFile();
		$result=$EpowerFile->getversioninfo($version);
		if($result['code']!=0){
			$this->ajaxReturn($result);
		}
		$files = $result['data']['content'];
		foreach ($files as &$file) {
			if($file['hash'] == 'UNLINK'){
				unset($file['old_hash']);
				continue;
			}
			if($file['old_hash']){
				$file['modify']=$file['old_hash'] != sha1_file(REAL_PATH . $file['url']);
			}else{
				$file['modify']=false;
			}
			unset($file['old_hash']);
		}
		$result['data']['content'] = $files;
		$this->ajaxReturn($result);
	}

	public function connect($version)
	{
		$EpowerFile=new EpowerFile();
		//下载升级包
		$result=$EpowerFile->download($version);
		if(empty($result)){
			$this->ajaxReturn(['code'=>9999, 'message'=>'升级文件下载失败！']);
		}

		//建立升级目录
		$version_dir = $this->dataDir . '/versions/' . $version;
		if(!is_dir($version_dir)){
			if(mkdir($version_dir, 0755, true) === false){
				$this->ajaxReturn(['code'=>9999, 'message'=>'升级目录创建失败！']);
			}
		}

		//解包
		$filename= $version_dir . '/version.tar';
		if(file_put_contents($filename, $result) === false){
			$this->ajaxReturn(['code'=>9999, 'message'=>'升级文件下载失败，请检查/datas/upgrade目录的写入权限。']);
		} 
		try{
			TarReader::unpack($filename, $version_dir . '/');

			//升级包版本判断
			$version_file = $version_dir . '/version.json';
			$hash_file = $version_dir . '/version.hash';
			$sql_file = $version_dir . '/version.sql';
			$json = file_get_contents($version_file);
			if(!$json){
				$this->ajaxReturn(['code'=>9999, 'message'=>'升级包错误！']);
			}
			$config=json_decode($json,true);

			$sql = is_file($sql_file);
			$content = file_get_contents($hash_file);
			$content = str_replace("\r\n","\n",$content);
			$content = str_replace("\r","\n",$content);
			$array=explode("\n",$content);

			$files = [];
			//读出升级包中的关联文件
			foreach ($array as $v) {
				$file =explode(":",$v);
				if(count($file) != 2){
					continue;
				}
				$files[$file[1]] = $file[0];
			}

			//备份文件
			$backup = date('YmdHis');
			$this->backup($files, $backup);
			//写文件
			foreach ($files as $url => $hash) {
				$target = REAL_PATH . $url;
				if($hash == 'UNLINK' && is_file($target)){
					unlink($target);
					continue;
				}

				$source = $version_dir . '/files';

				$dirname = dirname($target);
				if(!is_dir($dirname)){
					mkdir($dirname, 0755, true);
				}
				if(!is_file($source . $url)){
					continue;
				}

				//判断文件，防止空文件覆盖有内容的文件
				$contents = file_get_contents($source . $url);
				if(strlen($contents)==0){
					continue;
				}
				if(file_put_contents($target, $contents)===false){
					$this->rmdir($version_dir);
					$this->ajaxReturn(['code'=>9999, 'message'=>'文件写入失败！']);
				}
			}
			$config['update_time'] = date('Y-m-d H:i:s');
			if(file_put_contents($this->dataDir . '/data/version.json', json_encode($config))===false){
				$this->rmdir($version_dir);
				$this->ajaxReturn(['code'=>9999, 'message'=>'文件写入失败！']);
			}
			
			$result=true;
			if($sql){
				M()->startTrans();
				$db=M()->db();
				try{
					$result = $this->create_tables($db, file_get_contents($sql_file));
				}catch(\Exception $ex){
					$result = false;
				}
				if($result === false){
					M()->rollback();
					$this->rmdir($version_dir);
					$this->ajaxReturn(['code'=>9999, 'message'=>'数据库升级失败']);
				}else{
					M()->commit();
				}
			}
			$this->rmdir($version_dir);
			$this->ajaxReturn(['code'=>0, 'message'=>'升级成功！', 'backup' => $backup]);
		}catch(\Exception $ex){
			$this->rmdir($version_dir);
			$this->ajaxReturn(['code'=>9999, 'message'=>'升级错误']);
		}
	}
	private function backup($files, $rnd){
		$backupdir = $this->dataDir . '/backup/' . $rnd;
		if(!is_dir($backupdir)){
			mkdir($backupdir, 777, true);
		}
		foreach ($files as $url => $hash) {
			$target = $backupdir . $url;
			$dirname = dirname($target);
			if(!is_dir($dirname)){
				mkdir($dirname, 777, true);
			}
			copy(REAL_PATH . $url ,$target);
		}
	}
	private function rmdir($dir){
		$this->cleardir($dir);
	}
	public function download($version)
	{	
		if(!$version){
			$this->error("请先检查版本");
		}
		$EpowerFile=new EpowerFile();
		$a=$EpowerFile->download($version);
		header('Content-disposition: attachment; filename='.basename($version.".tar")); //文件名  
		header("Content-Type: application/tar"); //zip格式的  
		echo $a;
	}
	public function showfile($version)
	{	
		if(!$version){
			$this->error("请先检查版本");
		}
		$EpowerFile=new EpowerFile();
		$a=$EpowerFile->showfile($version, I('get.hash'), I('get.type'));
		$ext = I('get.ext');
		if(empty($ext)){
			$ext = '.php';
		}
		switch ( $ext )
		{
			case '.php':
			case '.css':
			case '.html':
			case '.js':
			case '.sql':
				header("Content-Type: text/html");
				echo '<pre>' . str_replace('>', '&gt;', str_replace('<', '&lt;', $a)) . '</pre>';
				break;
			case '.bmp':
			case '.gif':
			case '.png':
				header("Content-Type: image/" . substr($ext, 1));
				echo $a;
				break;
			case '.jpg':
				header("Content-Type: image/jpeg");
				echo $a;
				break;
			default:
				echo $a;
				break;
		}
	}
	
	private function cleardir($dirName){
		$handle = opendir($dirName);
		if(!$handle){
			return;
		}
		$path = '';
		while (false !== ($item = readdir($handle))) {
			if ($item == "." || $item == "..") {
				continue;
			}
			$path = "$dirName/$item";
			if (is_dir($path)) {
				$this->cleardir($path);
				continue;
			}
			unlink($path) ;
		}
		closedir($handle);
		rmdir($dirName);
	}
	
	/**
	 * 创建数据表
	 * @param  resource $db 数据库连接资源
	 */
	public function create_tables($db, $sql){
		//读取SQL文件

		$sql = "SET FOREIGN_KEY_CHECKS=0;\n" . $sql . ";\nSET FOREIGN_KEY_CHECKS=1";
		$sql = str_replace("\r\n", "\n", $sql);
		$sql = str_replace("\r", "\n", $sql);
		$sql = preg_replace('/^\-\-(.*?)$/m','', $sql);
		$sql = preg_replace('/\n+/',"\n",$sql);
		$sql = explode(";\n", $sql);
		foreach ($sql as $value) {
			$value = trim($value);
			if(empty($value)) continue;
			if($db->execute($value) === false){
				return false;
			}
		}
		return true;
	}
}