<?php
namespace EB\Tar;
use EB\Tar\TarHeader; 


/*
* @DESCRIPTION	tar file reader
*/	
class TarReader{
	private $desttar = null;

	/*
	* @DESCRIPTION	__construct function
	* @PARAM	$dest	: tar file that need to read
	*/	
	public function __construct($dest, $gzip=false){
		if($gzip === true){
			$dest = 'compress.zlib://' . $dest;
		}
		$fp = fopen($dest, 'r');
		if(!$fp){
			throw new \Exception('can not open target file');
		}
		$this->desttar = $fp;
		if($gzip!==true){
			$stat = fstat($fp);
			$size = $stat['size'];
			if($size<1536){
				throw new \Exception('tar file error');
			}
			fseek($fp, $size-512);
			$tail = $this->read_fixed(512);
			$tail = unpack('c*', $tail);
			
			if(array_sum($tail)!=0){
				throw new \Exception('tar file error 2');
			}
		}
		fseek($fp, 0);
	}

	/*
	* @DESCRIPTION	close tar
	*/
	public function close(){
		$this->lastfile = null;
		fclose($this->desttar);
	}

	/*
	* @DESCRIPTION	static methodunpack tar file to dir
	* @PARAM	$tarfile	: tar file path
	* @PARAM	$dest	: dest dir
	*/
	public static function unpack($tarfile, $dest, $gzip=false){
		if($gzip===false && pathinfo($tarfile, PATHINFO_EXTENSION ) == 'gz'){
			$gzip = true;
		}
		$tar = new TarReader($tarfile, $gzip);
		while($header = $tar->next()){
			if($header->isdir()){
				TarReader::create_dir($dest . $header->get_filename());
			}else{
				$tar->save($header, $dest . $header->get_filename());
			}
		}
		$tar->close();
	}
	
	private $lastfile = null;

	/*
	* @DESCRIPTION	read next file, return NULL at eof
	*/	
	public function next(){
		if($this->lastfile!=null){
			$size = $this->lastfile->get_filesize();
			$size = ceil($size / 512) * 512;
			fseek($this->desttar, $this->lastfile->get_offset() + $size);
		}
		$this->lastfile = null;
		$result = $this->read_fixed(512);
		if($result[0] == "\0"){
			return null;
		}
		$filename = str_replace("\0","",substr($result, 0, 100));
		$prefix = str_replace("\0","",substr($result, -167));
		if(!empty($prefix)){
			$filename = $prefix . '\\' . $filename;
		}

		$filesize = trim(substr($result, 124, 12));
		$filesize = intval($filesize, 8);
		$tag = unpack('n', substr($result, 100, 2));
		$header = new TarHeader($filename, $tag[1] == 0x2034);
		$header->set_filesize($filesize);
		$header->set_offset(ftell($this->desttar));
		$this->lastfile = $header;
		return $header;
	}

	/*
	* @DESCRIPTION	save single file
	* @PARAM	$header	: file info
	* @PARAM	$dest	: dest file path
	*/
	public function save($header, $dest){
		TarReader::create_dir(dirname($dest));
		fseek($this->desttar, $header->get_offset());
		$dest = str_replace('\\', '/', $dest);
		$dest = str_replace('\\\\', '/', $dest);
		$fp = fopen($dest, 'w');
		if(!$fp){
			return false;
		}
		try{
			$left = $header->get_filesize();
			$blocksize = 8192;
			while($left>0){
				if($left<8192){
					$blocksize = $left;
				}
				$result = fread($this->desttar, $blocksize);
				fwrite($fp, $result);
				$left -= strlen($result);
			}
		}catch(\Exception $e){
			fseek($this->desttar, $header->get_offset());
			return false;
		}finally{
			fclose($fp);
		}
	}

	/*
	* @DESCRIPTION	get file contents
	* @PARAM	$header	: file info
	*/
	public function get_contents($header){
		fseek($this->desttar, $header->get_offset());
		try{
			$contents = '';
			$left = $header->get_filesize();
			$blocksize = 8192;
			while($left>0){
				if($left<8192){
					$blocksize = $left;
				}
				$result = fread($this->desttar, $blocksize);
				$contents .= $result;
				$left -= strlen($result);
			}
			return $contents;
		}catch(\Exception $e){
			fseek($this->desttar, $header->get_offset());
			return false;
		}
	}

	private function read_fixed($length){
		$left = $length;
		$result = '';
		while($left>0 && !feof($this->desttar)){
			$result .= fread($this->desttar, $left);
			$left = $length - strlen($result);
		}
		return $result;
	}

	/*
	* @DESCRIPTION	static method, create dir. auto create parent dir
	* @PARAM	$dir	: dest dir
	*/
	public static function create_dir($dir){
		if(is_dir($dir)){
			return;
		}
		mkdir($dir, 0777, true);
	}
}
